/*
 * txt2fb - ASCII Text Filter for Framebuffer Device
 * Copyright (C)2013 Finalmedia Joerg Reinhard e.K. (www.finalmedia.de)
 *
 * Source code is released under the modified BSD license.
 *
 * This FILTER is reading ASCII-Text from stdin, converting bytes in bit sequences
 * and writing 32bit RGBA framebuffer-data to stdout,
 * (I really mean stdout and I do not fopen directly a specific /dev/fb!)
 *
 * So you can simply redirect the output from stdout to /dev/fb0, /dev/fb1 ... /dev/fb_n.
 * Advantages: You can easyly use an unnamed pipe and "tee" to get a kind of "dualscreen-hack":
 *
 *     ./txt2fb 1024 768 < README | tee /dev/fb0 > /dev/fb1
 *
 * You can also save the stream to a file, crypt the stream with aespipe,
 * or pipe it trough network with netcat to a remote host,
 * even tunnel trough ssh or something like that.
 *
 * 8x7 Font is hardcoded. Finally a Glyph is 8x8 Pixel because of
 * additional horizontal 1 px spacing-line per char.
 *
 * Example: ./txt2fb 1024 768 < README > /dev/fb0
 *
 */

#include <stdio.h>

int main(int argc, char *argv[])
{

/* at Screen-Resolution 1280x1024 px is 160x128, 1024x768 is 128x96 - because of 8x8 blocks) */
int char_per_line=128,max_lines=96,bits_per_pixel=32;

if (argc<3) {
fprintf(stderr,"example: txt2fb 1024 768 32 < README > /dev/fb0\n");
return -1;
} else {
char_per_line=strtol(argv[1],(char **) NULL,10)/8;  /* 128, if resolution 1024x768 */
max_lines=strtol(argv[2],(char **) NULL,10)/8;      /* 96,  if resolution 1024x768 */
bits_per_pixel=strtol(argv[3],(char **) NULL,10);	  /* 8,16 oder 32 bit */
}

if (char_per_line>1024 || char_per_line < 1 || max_lines < 1 || max_lines > 1024)
{
fprintf(stderr,"Sorry. Screen resolution has to be minimal 8x8 pixels and maximal 8192x8192 pixels.\n");
return -1;
}

/* Allocate enough memory for 1024 chars per line, 8192 pixel */
char charzeile[1024];

/*
 * Finalmedia Bitmap Font ;) Nothing special, I know
 * And i didn'nt had the the time to draw small letters
 * 94 Chars in a 8x7 Font Matrix (have a look at README)
 * Readable ASCII Chars, octal 32 to 126
 *
*/
int shift=32;
char font[95][7] = {
	{ 0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  /*   (032) 000 */
	{ 0x08,0x08,0x08,0x08,0x08,0x00,0x08 },  /* ! (033) 001 */
	{ 0x66,0x66,0x00,0x00,0x00,0x00,0x00 },  /* " (034) 002 */
	{ 0x24,0x24,0xFF,0x24,0xFF,0x24,0x24 },  /* # (035) 003 */
	{ 0x08,0x3E,0x28,0x3E,0x0A,0x3E,0x08 },  /* $ (036) 004 */
	{ 0x01,0x62,0x64,0x08,0x13,0x23,0x40 },  /* % (037) 005 */
	{ 0x1C,0x22,0x20,0x1D,0x42,0x42,0x3D },  /* & (038) 006 */
	{ 0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  /*   (039) 007 */
	{ 0x02,0x04,0x08,0x08,0x08,0x04,0x02 },  /* ( (040) 008 */
	{ 0x40,0x20,0x10,0x10,0x10,0x20,0x40 },  /* ) (041) 009 */
	{ 0x00,0x49,0x2A,0x1C,0x2A,0x49,0x00 },  /* * (042) 010 */
	{ 0x00,0x08,0x08,0x3E,0x08,0x08,0x00 },  /* + (043) 011 */
	{ 0x00,0x00,0x00,0x00,0x60,0x20,0x40 },  /* , (044) 012 */
	{ 0x00,0x00,0x00,0x7E,0x00,0x00,0x00 },  /* - (045) 013 */
	{ 0x00,0x00,0x00,0x00,0x00,0x60,0x60 },  /* . (046) 014 */
	{ 0x01,0x02,0x04,0x08,0x10,0x20,0x40 },  /* / (047) 015 */
	{ 0x3C,0x42,0x46,0x4A,0x52,0x62,0x3C },  /* 0 (048) 016 */
	{ 0x06,0x0A,0x12,0x02,0x02,0x02,0x02 },  /* 1 (049) 017 */
	{ 0x7E,0x02,0x02,0x7E,0x40,0x40,0x7E },  /* 2 (050) 018 */
	{ 0x7E,0x02,0x02,0x7E,0x02,0x02,0x7E },  /* 3 (051) 019 */
	{ 0x42,0x42,0x42,0x7E,0x02,0x02,0x02 },  /* 4 (052) 020 */
	{ 0x7E,0x40,0x40,0x7E,0x02,0x02,0x7E },  /* 5 (053) 021 */
	{ 0x7E,0x40,0x40,0x7E,0x42,0x42,0x7E },  /* 6 (054) 022 */
	{ 0x7E,0x02,0x04,0x08,0x10,0x20,0x40 },  /* 7 (055) 023 */
	{ 0x7E,0x42,0x42,0x7E,0x42,0x42,0x7E },  /* 8 (056) 024 */
	{ 0x7E,0x42,0x42,0x7E,0x02,0x02,0x7E },  /* 9 (057) 025 */
	{ 0x00,0x00,0x18,0x00,0x18,0x00,0x00 },  /* : (058) 026 */
	{ 0x00,0x00,0x18,0x00,0x18,0x08,0x10 },  /* ; (059) 027 */
	{ 0x04,0x08,0x10,0x20,0x10,0x08,0x04 },  /* < (060) 028 */
	{ 0x00,0x00,0x7E,0x00,0x7E,0x00,0x00 },  /* = (061) 029 */
	{ 0x20,0x10,0x08,0x04,0x08,0x10,0x20 },  /* > (062) 030 */
	{ 0x3E,0x41,0x02,0x04,0x08,0x00,0x08 },  /* ? (063) 031 */
	{ 0x3E,0xC3,0x99,0xA5,0xA6,0xD8,0x7E },  /* @ (064) 032 */
	{ 0x3C,0x42,0x42,0x7E,0x42,0x42,0x42 },  /* A (065) 033 */
	{ 0x7C,0x42,0x42,0x7C,0x42,0x42,0x7C },  /* B (066) 034 */
	{ 0x7E,0x40,0x40,0x40,0x40,0x40,0x7E },  /* C (067) 035 */
	{ 0x7C,0x42,0x42,0x42,0x42,0x42,0x7C },  /* D (068) 036 */
	{ 0x7E,0x40,0x40,0x7E,0x40,0x40,0x7E },  /* E (069) 037 */
	{ 0x7E,0x40,0x40,0x7E,0x40,0x40,0x40 },  /* F (070) 038 */
	{ 0x7E,0x40,0x40,0x5E,0x42,0x42,0x7E },  /* G (071) 039 */
	{ 0x42,0x42,0x42,0x7E,0x42,0x42,0x42 },  /* H (072) 040 */
	{ 0x10,0x10,0x10,0x10,0x10,0x10,0x10 },  /* I (073) 041 */
	{ 0x02,0x02,0x02,0x02,0x02,0x02,0x3C },  /* J (074) 042 */
	{ 0x42,0x44,0x48,0x70,0x48,0x44,0x42 },  /* K (075) 043 */
	{ 0x40,0x40,0x40,0x40,0x40,0x40,0x7E },  /* L (076) 044 */
	{ 0x42,0x66,0x5A,0x42,0x42,0x42,0x42 },  /* M (077) 045 */
	{ 0x41,0x61,0x51,0x49,0x45,0x43,0x41 },  /* N (078) 046 */
	{ 0x3C,0x42,0x42,0x42,0x42,0x42,0x3C },  /* O (079) 047 */
	{ 0x7F,0x41,0x41,0x7F,0x40,0x40,0x40 },  /* P (080) 048 */
	{ 0x3C,0x42,0x42,0x42,0x42,0x42,0x3D },  /* Q (081) 049 */
	{ 0x7C,0x42,0x42,0x7C,0x44,0x42,0x41 },  /* R (082) 050 */
	{ 0x3E,0x40,0x40,0x3C,0x02,0x02,0x7C },  /* S (083) 051 */
	{ 0x7F,0x08,0x08,0x08,0x08,0x08,0x08 },  /* T (084) 052 */
	{ 0x41,0x41,0x41,0x41,0x41,0x41,0x3E },  /* U (085) 053 */
	{ 0x41,0x41,0x41,0x41,0x22,0x14,0x08 },  /* V (086) 054 */
	{ 0x41,0x49,0x49,0x49,0x49,0x49,0x36 },  /* W (087) 055 */
	{ 0x41,0x22,0x14,0x1C,0x14,0x22,0x41 },  /* X (088) 056 */
	{ 0x41,0x22,0x14,0x08,0x08,0x08,0x08 },  /* Y (089) 057 */
	{ 0x7F,0x02,0x04,0x08,0x10,0x20,0x7F },  /* Z (090) 058 */
	{ 0x0F,0x08,0x08,0x08,0x08,0x08,0x0F },  /* [ (091) 059 */
	{ 0x80,0x40,0x28,0x10,0x08,0x04,0x02 },  /* \ (092) 060 */
	{ 0xF8,0x08,0x08,0x08,0x08,0x08,0xF8 },  /* ] (093) 061 */
	{ 0x00,0x14,0x22,0x00,0x00,0x00,0x00 },  /* ^ (094) 062 */
	{ 0x00,0x00,0x00,0x00,0x00,0x00,0xFF },  /* _ (095) 063 */
	{ 0x80,0x20,0x00,0x00,0x00,0x00,0x00 },  /* ` (096) 064 */
	{ 0x3C,0x42,0x42,0x7E,0x42,0x42,0x42 },  /* a (097) 065 */
	{ 0x7C,0x42,0x42,0x7C,0x42,0x42,0x7C },  /* b (098) 066 */
	{ 0x7E,0x40,0x40,0x40,0x40,0x40,0x7E },  /* c (099) 067 */
	{ 0x7C,0x42,0x42,0x42,0x42,0x42,0x7C },  /* d (100) 068 */
	{ 0x7E,0x40,0x40,0x7E,0x40,0x40,0x7E },  /* e (101) 069 */
	{ 0x7E,0x40,0x40,0x7E,0x40,0x40,0x40 },  /* f (102) 070 */
	{ 0x7E,0x40,0x40,0x5E,0x42,0x42,0x7E },  /* g (103) 071 */
	{ 0x42,0x42,0x42,0x7E,0x42,0x42,0x42 },  /* h (104) 072 */
	{ 0x10,0x10,0x10,0x10,0x10,0x10,0x10 },  /* i (105) 073 */
	{ 0x02,0x02,0x02,0x02,0x02,0x02,0x3C },  /* j (106) 074 */
	{ 0x42,0x44,0x48,0x70,0x48,0x44,0x42 },  /* k (107) 075 */
	{ 0x40,0x40,0x40,0x40,0x40,0x40,0x7E },  /* l (108) 076 */
	{ 0x42,0x66,0x5A,0x42,0x42,0x42,0x42 },  /* m (109) 077 */
	{ 0x41,0x61,0x51,0x49,0x45,0x43,0x41 },  /* n (110) 078 */
	{ 0x3C,0x42,0x42,0x42,0x42,0x42,0x3C },  /* o (111) 079 */
	{ 0x7F,0x41,0x41,0x7F,0x40,0x40,0x40 },  /* p (112) 080 */
	{ 0x3C,0x42,0x42,0x42,0x42,0x42,0x3D },  /* q (113) 081 */
	{ 0x7C,0x42,0x42,0x7C,0x44,0x42,0x41 },  /* r (114) 082 */
	{ 0x3E,0x40,0x40,0x3C,0x02,0x02,0x7C },  /* s (115) 083 */
	{ 0x7F,0x08,0x08,0x08,0x08,0x08,0x08 },  /* t (116) 084 */
	{ 0x41,0x41,0x41,0x41,0x41,0x41,0x3E },  /* u (117) 085 */
	{ 0x41,0x41,0x41,0x41,0x22,0x14,0x08 },  /* v (118) 086 */
	{ 0x41,0x49,0x49,0x49,0x49,0x49,0x36 },  /* w (119) 087 */
	{ 0x41,0x22,0x14,0x1C,0x14,0x22,0x41 },  /* x (120) 088 */
	{ 0x41,0x22,0x14,0x08,0x08,0x08,0x08 },  /* y (121) 089 */
	{ 0x7F,0x02,0x04,0x08,0x10,0x20,0x7F },  /* z (122) 090 */
	{ 0x06,0x08,0x08,0x70,0x08,0x08,0x06 },  /* { (123) 091 */
	{ 0x18,0x18,0x18,0x18,0x18,0x18,0x18 },  /* | (124) 092 */
	{ 0x60,0x10,0x10,0x0E,0x10,0x10,0x60 },  /* } (125) 093 */
	{ 0x00,0x00,0x60,0x99,0x06,0x00,0x70 }   /* ~ (126) 094 */
	};


char byte;
char bits;
char ch;
char input_ch;
char value,fill;

short buffer=0;
short charcount=0;
short linecount=0;
short lines=0;

int k;

while(1) {

	input_ch=getc(stdin);

	/* buffer current character, if valid */
	if (input_ch>31 && input_ch<127 && input_ch!='\n') { charzeile[buffer]=input_ch; buffer++; }

	/* if linebreak detected, fill buffer with empty space until end of line */
	if (input_ch=='\n') { for (k=buffer;k<char_per_line+1;k++) { charzeile[k]=' '; }
	buffer=char_per_line+1; lines--; }

	/* if buffer is full, output to stdout in framebuffer-format */
	if ((buffer>char_per_line)) {

		for(linecount=0; linecount<7; linecount++) {
			for(charcount=0; charcount<char_per_line; charcount++) {
				ch=charzeile[charcount]; /* current char in buffer */

				fill=0xFF;
				/* if char is a number, than draw it light gray ;) */
				if (ch-shift>=16 && ch-shift<26) fill=0xCC;

				byte=font[(ch-shift)][linecount]; /* value of current character in matrix, in actual scanline, 8bit */

				/* read single bits of byte as a sequence */
				for(bits=0;bits<8;bits++) {
				if ((byte<<bits)&0x80) value=fill; else value=0x00;

				putc(value,stdout);  /* R ? */
				putc(value,stdout);  /* G ? */
				if (bits_per_pixel>23) { putc(value,stdout); }  /* B ? */
				if (bits_per_pixel>31) { putc(0xFF,stdout); }  /* Alpha ? */
				}
			}

		}


		/* after 7 lines add another line of empty space (additional pixel line to blow up 7 lines to 8) */
		for(k=0;k<(char_per_line*8);k++) {
		putc(0x00,stdout);
		putc(0x00,stdout);
		if (bits_per_pixel>23) { putc(0x00,stdout); }
		if (bits_per_pixel>31) { putc(0xFF,stdout); }
		}

		/* if current line is written, reset buffer and switch to next line */
		buffer=0;
		lines++;
		fflush(stdout);


	}

if (lines>max_lines-1 || input_ch==EOF) { return 0; }


}
}
